import json

from django import forms
from django.core.exceptions import ImproperlyConfigured

from galleryfield import conf, defaults
from galleryfield.utils import (convert_dict_to_plain_text,
                                get_formatted_thumbnail_size, get_url_from_str,
                                logger)


class GalleryWidget(forms.HiddenInput):
    """This is the default widget used by :class:`galleryfield.fields.GalleryFormField`.

    :param upload_url: An URL name or an URL of the upload handler
           view used by the widget instance, defaults to `None`.
           When not set, this param will be auto-configured if the parent
           :class:`galleryfield.fields.GalleryFormField` is used by
           a :class:`galleryfield.fields.GalleryField`, and the value will follow
           the :ref:`naming rule <image_handling_url_naming_rule>`. 
           The value can also be `None` if set explicitly by::
           
               self.fields["images"].widget.upload_url = None

           In this case, upload ui won't show upload buttons.
    :type upload_url: str, optional
    :param fetch_url: An URL name or an URL for fetching the existing
           images in the gallery instance, defaults to `None`.
           When not set, this param will be auto-configured if the parent
           :class:`galleryfield.fields.GalleryFormField` is used by
           a :class:`galleryfield.fields.GalleryField`, the value wil. follow
           the :ref:`naming rule <image_handling_url_naming_rule>`.
    :type fetch_url: str, optional
    :param multiple: Whether allow to select multiple image files in the 
           file picker. Defaults to ``True``.
    :type multiple: bool, optional
    :param thumbnail_size: The thumbnail size (both width and height), defaults 
           to ``defaults.DEFAULT_THUMBNAIL_SIZE``, which can be overridden by
           ``settings.DJANGO_GALLERY_FIELD_CONFIG["thumbnails"]["size"]``.
           The value can be set after the widget is initialized.
    :type thumbnail_size: int, optional
    :param template: The path of template which is used to render the widget.
           defaults to ``galleryfield/widget.html``, which support template 
           inheritance.
    :type template: str, optional
    :param attrs: Html attribute when rendering the field (Which is a 
           :class:`django.forms.HiddenInput`), defaults to `None`. See 
           `Django docs <https://docs.djangoproject.com/en/dev/ref/forms/widgets/#django.forms.Widget.attrs>`_.
    :type attrs: dict, optional
    :param options: Other options when rendering the widget. Implemented options:

           * **accepted_mime_types** (`list`, `optional`) - A list of MIME types
             used to filter files when picking files with file picker, defaults to ``['image/*']``
    :type options: dict, optional
    :param jquery_file_upload_ui_options: The default template is using 
           blueimp/jQuery-File-Upload package to render the ui and dealing with
           AJAX upload. See :setting:`jquery_file_upload_ui_options` for more information.
    :type jquery_file_upload_ui_options: dict, optional
    :param disable_fetch: Whether disable fetching existing images of the
           form instance (if any), defaults to `False`. If True, the validity of
           ``fetch_url`` will not be checked.
    :type disable_fetch: bool, optional
    :param disable_server_side_crop: Whether disable server side cropping of 
           uploaded images, defaults to `False`.
    :type disable_server_side_crop: bool, optional
    
    """  # noqa

    def __init__(
            self,
            upload_url=None,
            fetch_url=None,
            multiple=True,
            thumbnail_size=conf.DEFAULT_THUMBNAIL_SIZE,
            template="galleryfield/widget.html",
            attrs=None, options=None,
            jquery_file_upload_ui_options=None,
            disable_fetch=False,
            disable_server_side_crop=False,
            **kwargs):

        super(GalleryWidget, self).__init__(attrs)

        self.multiple = multiple
        self._thumbnail_size = get_formatted_thumbnail_size(thumbnail_size)
        self.template = template

        self.disable_fetch = disable_fetch
        self.disable_server_side_crop = disable_server_side_crop

        self.upload_url = upload_url

        self.fetch_url = (
            None if disable_fetch else fetch_url)

        self.jquery_file_upload_ui_options = jquery_file_upload_ui_options or {}

        self.options = options and options.copy() or {}
        self.options.setdefault("accepted_mime_types", ['image/*'])

    @property
    def thumbnail_size(self):
        return self._thumbnail_size

    @thumbnail_size.setter
    def thumbnail_size(self, value):
        self._thumbnail_size = get_formatted_thumbnail_size(value)

    @property
    def jquery_file_upload_ui_options(self):
        return (self._jquery_file_upload_ui_options
                or conf.JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS)

    @jquery_file_upload_ui_options.setter
    def jquery_file_upload_ui_options(self, options):
        if options is None:
            return
        if not isinstance(options, dict):
            raise ImproperlyConfigured(
                "%(obj)s: 'jquery_file_upload_ui_options' must be a dict" % {
                    "obj": self.__class__.__name__
                }
            )
        ju_settings = (
            conf.JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS.copy())
        ju_settings.update(options)

        if "maxNumberOfFiles" in ju_settings:
            logger.warning(
                "%(obj)s: 'maxNumberOfFiles' in 'jquery_file_upload_ui_options' "
                "will be overridden later by the formfield. You should set that "
                "value in the formfield it belongs to, e.g. \n"
                "self.fields['my_gallery_field'].max_number_of_images = %(value)s"
                % {"obj": self.__class__.__name__,
                   "value": str(ju_settings["maxNumberOfFiles"])}
            )

        if ("singleFileUploads" in ju_settings
                and str(ju_settings["singleFileUploads"]).lower() == "false"):
            logger.warning(
                "%(obj)s: 'singleFileUploads=False' in "
                "'jquery_file_upload_ui_options' is not allowed and will be "
                "ignored."
                % {"obj": self.__class__.__name__}
            )

        if "previewMaxWidth" in ju_settings or "previewMaxHeight" in ju_settings:
            logger.warning(
                "%(obj)s: 'previewMaxWidth' and 'previewMaxHeight' in "
                "'jquery_file_upload_ui_options' are ignored. You should set "
                "the value by the 'thumbnail_size' option, e.g., "
                "thumbnail_size='120x60'"
                % {"obj": self.__class__.__name__}
            )

        self._jquery_file_upload_ui_options = ju_settings

    def set_and_check_urls(self):
        # We now then the URL names into an actual URL, that
        # can't be down in __init__ because url_conf is not
        # loaded when doing the system check, and will result
        # in failure to start.
        try:
            self.upload_url = get_url_from_str(
                    self.upload_url, require_urlconf_ready=True)
        except Exception:
            raise ImproperlyConfigured(
                "'upload_url' is invalid: %s is neither "
                "a valid URL nor a valid URL name."
                % self.upload_url)

        if self.disable_fetch:
            self.fetch_url = None
        else:
            try:
                self.fetch_url = get_url_from_str(
                        self.fetch_url, require_urlconf_ready=True)

            except Exception:
                raise ImproperlyConfigured(
                    "'fetch_url' is invalid: %s is neither "
                    "a valid URL nor a valid URL name."
                    % self.fetch_url)

        # In the following we validate update the urls from url names (if it is
        # not an url) and check the potential conflicts of init params
        # of the widget with the target_image_model set to the widget.
        # For validation of urls, because those urls are assigned by reverse_lazy,
        # the result of which will only be validated until evaluation, we
        # can only do that until the request context is available.
        # For checking conflicts, the logic is: if the target_image_model is not
        # using the built-in model, i.e., defaults.DEFAULT_TARGET_IMAGE_MODEL,
        # then the upload, crop and fetch views should not use the built-in ones.
        # Because those views are using defaults.DEFAULT_TARGET_IMAGE_MODEL
        # as the target image model.
        # BTW, this check can't be done at the model/field check stage or formfield
        # init stage, because we should allow the widget be changed after the form
        # is initialized. So, we have to do this before rendering the widget,
        # and throw the error if any.

        target_image_model = getattr(self, "image_model", None)

        image_model_is_default = (
                target_image_model == defaults.DEFAULT_TARGET_IMAGE_MODEL)

        if image_model_is_default:
            return

        conflict_config = []
        upload_url_is_default = (
            self.upload_url
            == get_url_from_str(defaults.DEFAULT_UPLOAD_URL_NAME,
                                require_urlconf_ready=True))

        if upload_url_is_default:
            conflict_config.append(
                {"param": "upload_url",
                 "value": self.upload_url})

        if self.fetch_url:
            fetch_url_is_default = (
                self.fetch_url
                == get_url_from_str(defaults.DEFAULT_FETCH_URL_NAME,
                                    require_urlconf_ready=True))
            if fetch_url_is_default:
                conflict_config.append(
                    {"param": "fetch_url",
                     "value": self.fetch_url})

        if not conflict_config:
            return

        widget_is_servicing = getattr(self, "widget_is_servicing")
        msgs = ["'%(obj)s' is using '%(used_model)s' "
                "instead of built-in '%(default_model)s', while "
                % {"obj": widget_is_servicing,
                   "used_model": target_image_model,
                   "default_model": defaults.DEFAULT_TARGET_IMAGE_MODEL}
                ]

        for cc in conflict_config:
            msgs.append(
                "'%(param)s' is using built-in value '%(value)s', " % cc)

        msgs.append(
            "which use built-in '%(default_model)s'. You need to write "
            "your own views for your image model." % {
                "default_model": defaults.DEFAULT_TARGET_IMAGE_MODEL
            })

        msg = "".join(msgs)
        raise ImproperlyConfigured(msg)

    class Media:
        js = tuple(conf.JS)
        css = {'all': tuple(conf.CSS)}

    @property
    def is_hidden(self):
        return False

    def get_stringfied_jquery_file_upload_ui_options(self):
        # See blueimp/jQuery-File-Upload
        # https://github.com/blueimp/jQuery-File-Upload/wiki/Options

        # We copy the options as the actual context used in rendering
        # so as to avoid logger warnings
        ui_options = self.jquery_file_upload_ui_options.copy()

        # Remove the option (i.e., use default False)
        ui_options.pop("singleFileUploads", None)

        # Remove other options which we are using default but
        # don't allow user to change
        for option in ["fileInput", "formData"]:
            ui_options.pop(option, None)

        # Fixme: this is hardcoded
        ui_options["paramName"] = "'files[]'"

        # override maxNumberOfFiles
        ui_options.pop("maxNumberOfFiles", None)
        max_number_of_images = (
            getattr(self, "max_number_of_images", None))

        if max_number_of_images:
            ui_options["maxNumberOfFiles"] = max_number_of_images

        # override previewMaxWidth and previewMaxHeight

        _width, _height = self.thumbnail_size.split("x")
        ui_options.update(
            {"previewMaxWidth": _width,
             "previewMaxHeight": _height,

             # This is used as a CSS selector to fine the input field
             "hiddenFileInput": "'.%s'" % conf.FILES_FIELD_CLASS_NAME,
             })

        return convert_dict_to_plain_text(ui_options, indent=16)

    def render(self, name, value, attrs=None, renderer=None):
        self.set_and_check_urls()

        context = {
            'input_string': super().render(name, value, attrs, renderer),
            'name': name,
            'multiple': self.multiple and 1 or 0,
            'thumbnail_size': str(self.thumbnail_size),
            'prompt_alert_on_window_reload_if_changed':
                conf.PROMPT_ALERT_ON_WINDOW_RELOAD_IF_CHANGED
        }

        # Do not fill in empty value to hidden inputs
        if value:
            context["pks"] = json.loads(value)
            context["fetch_url"] = self.fetch_url

        _context = self.get_context(name, value, attrs)

        if (_context["widget"]["attrs"].get("disabled", False)
                or _context["widget"]["attrs"].get("readonly")):
            context["uploader_disabled"] = True
        else:
            context.update({
                "upload_url": self.upload_url,
                "accepted_mime_types": self.options["accepted_mime_types"],
                "disable_server_side_crop": self.disable_server_side_crop
            })
        context["widget"] = _context["widget"]

        context["jquery_fileupload_ui_options"] = (
            self.get_stringfied_jquery_file_upload_ui_options())

        return renderer.render(template_name=self.template, context=context)



import logging

from django.apps import apps
from django.core.checks import Critical, Info
from django.core.exceptions import (AppRegistryNotReady, FieldDoesNotExist,
                                    ImproperlyConfigured)
from django.db.models import ImageField
from django.urls import (NoReverseMatch, Resolver404, resolve, reverse,
                         reverse_lazy)

from galleryfield import defaults

logger = logging.getLogger('django-galleryfield')


class DJGalleryCriticalCheckMessage(Critical):
    def __init__(self, *args, **kwargs):
        super(DJGalleryCriticalCheckMessage, self).__init__(*args, **kwargs)
        self.obj = self.obj or ImproperlyConfigured.__name__


INSTANCE_ERROR_PATTERN = "%(location)s must be an instance of %(types)s."
GENERIC_ERROR_PATTERN = "Error in %(location)s: %(error_type)s: %(error_str)s"
REQUIRED_CONF_ERROR_PATTERN = (
    "You must configure %(location)s for RELATE to run properly.")


def convert_dict_to_plain_text(d, indent=4):
    result = []
    for k, v in d.items():
        if v is not None:
            if v in [True, False]:
                v = str(v).lower()
            result.append(" " * indent + "%s: %s," % (k, str(v)))
    return "\n".join(result)


def get_or_check_image_field(obj, target_model, check_id_prefix, is_checking=False):
    """Get the image field from target image model, or check if the params will work
    in the model during model.check().

    :param target_model: a string or ``None``. If ``None``,
           ``defaults.DEFAULT_TARGET_IMAGE_MODEL`` will be used.
    :type target_model: str
    :param check_id_prefix: The function will be used in model checks and formfield
           checks . They will use different check message id.
    :param obj: Where the check error is detected, used for model checks.
    :param is_checking: whether this function is used in checks.
    :return: when ``is_checking`` is ``False``, we were getting the image field from
      the target image model. if any errors this will return ``None``, and the errors
      will be raised in model checks, i.e., when ``is_checking`` is ``True``.
    """

    errors = []
    target_model_klass = None
    will_proceed_checking_target_model_image_fields = True
    if target_model is not None:
        if not isinstance(target_model, str):
            errors.append(DJGalleryCriticalCheckMessage(
                msg=(INSTANCE_ERROR_PATTERN
                     % {"location": str(obj),
                        "types": "str"}),
                id="%s.E001" % check_id_prefix,
                obj=obj
            ))
            will_proceed_checking_target_model_image_fields = False
        else:
            try:
                target_model_klass = apps.get_model(target_model)
            except AppRegistryNotReady:
                # Preventing from failing in field initialization
                # We will raise that in field check
                assert not is_checking
                return None
            except Exception as e:
                if not is_checking:
                    return None

                # todo: add doc ref in the error
                errors.append(DJGalleryCriticalCheckMessage(
                    msg=(GENERIC_ERROR_PATTERN
                         % {"location": str(obj),
                            "error_type": type(e).__name__,
                            "error_str": str(e)}
                         + "\n See "
                           "https://docs.djangoproject.com/en/dev/ref/applications/#django.apps.AppConfig.get_model"  # noqa
                           " for more information."
                         ),
                    id="%s.E002" % check_id_prefix,
                    obj=obj
                ))
                will_proceed_checking_target_model_image_fields = False

    if will_proceed_checking_target_model_image_fields:
        if target_model_klass is None:
            target_model = defaults.DEFAULT_TARGET_IMAGE_MODEL
            try:
                target_model_klass = apps.get_model(target_model)
            except AppRegistryNotReady:
                assert not is_checking
                # Preventing from failing in field initialization
                # We will raise that in field check
                return None
            # Here we assume defaults.DEFAULT_TARGET_IMAGE_MODEL will always work
            else:
                errors.append(Info(
                    msg=('"target_model" is set to None, '
                         '"%(default)s" is used as default.'
                         % {"default": defaults.DEFAULT_TARGET_IMAGE_MODEL}
                         ),
                    id="%s.I001" % check_id_prefix,
                    obj=obj
                ))

        get_image_field_class_method = None
        get_image_field_class_method_raised_error = False
        try:
            image_field = (
                target_model_klass._meta.get_field(
                    defaults.DEFAULT_TARGET_IMAGE_FIELD_NAME))
            if type(image_field) is not ImageField:
                raise FieldDoesNotExist()
        except FieldDoesNotExist:
            image_field = None
            get_image_field_class_method = (
                getattr(target_model_klass, "get_image_field", None))
            if get_image_field_class_method is not None:
                if callable(get_image_field_class_method):
                    try:
                        image_field = get_image_field_class_method()
                    except Exception as e:
                        if not is_checking:
                            # Preventing from failing in field initialization
                            # We will raise that in field check
                            return None
                        errors.append(Critical(
                            msg=('Error in %(location)s: model %(model)s defined '
                                 '"get_image_field" method failed with'
                                 ' %(exception)s: %(str_e)s'
                                 % {"location": str(obj),
                                    "model": target_model,
                                    "exception": type(e).__name__,
                                    "str_e": str(e)
                                    }),
                            id="%s.E003" % check_id_prefix,
                            obj=obj
                        ))
                        get_image_field_class_method_raised_error = True

        if image_field is not None:
            if type(image_field) is not ImageField:
                image_field = None

        if image_field is None:
            if not is_checking:
                # Preventing from failing in field initialization
                # We will raise that in field check
                return None
            if get_image_field_class_method is None:
                errors.append(Critical(
                    msg=('Error in %(location)s: model %(model)s must '
                         'either have a field named "image" '
                         'or has a classmethod named "get_image_field", '
                         'which returns the image field of the model'
                         % {"location": str(obj),
                            "model": target_model
                            }),
                    id="%s.E004" % check_id_prefix,
                    obj=obj
                ))
            elif not get_image_field_class_method_raised_error:
                errors.append(Critical(
                    msg=('Error in %(location)s: model %(model)s defined '
                         '"get_image_field" class method '
                         'did not return a ImageField type'
                         % {"location": str(obj),
                            "model": target_model
                            }),
                    id="%s.E005" % check_id_prefix,
                    obj=obj
                ))
        else:
            if not is_checking:
                return image_field

    return errors


def get_url_from_str(url_str, require_urlconf_ready=False):
    """

    :param url_str:
    :param require_urlconf_ready: During the system loading stage,
     We had to use ``reverse_lazy``, or else it will result in
     circular import.
     Ref: https://docs.djangoproject.com/en/dev/ref/urlresolvers/#reverse-lazy
     And that means wen url_str will only be evaluated and validate
     before requests.
    :return: an URL string or a lazy reverse object
    """
    if not url_str:
        return None
    try:
        resolve(url_str)
    except Resolver404:
        if not require_urlconf_ready:
            return reverse_lazy(url_str)
        try:
            return reverse(url_str)
        except NoReverseMatch:
            raise ImproperlyConfigured(
                "'%s' is neither a valid URL nor a valid URL name" % url_str)
    return url_str


class InvalidThumbnailFormat(ValueError):
    pass


def get_formatted_thumbnail_size(thumbnail_size, name="thumbnail_size"):
    if not thumbnail_size:
        thumbnail_size = defaults.DEFAULT_THUMBNAIL_SIZE

    def get_thumb_size_from_iterator(size_iter):
        if not 0 < len(size_iter) <= 2:
            raise InvalidThumbnailFormat()
        for _item in size_iter:
            if not str(_item).isdigit():
                raise InvalidThumbnailFormat(str(_item))
        if len(size_iter) == 1:
            size_iter = size_iter * 2
        return "x".join(list(map(str, size_iter)))

    if not isinstance(thumbnail_size, (list, tuple)):
        thumbnail_size = str(thumbnail_size).strip()
        if not thumbnail_size:
            raise ValueError(
                "'%s' can't be an empty string" % name)
        thumbnail_size = [s.strip() for s in thumbnail_size.lower().split("x")]

    return get_thumb_size_from_iterator(thumbnail_size)



JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS = {
    "autoUpload": False,
    "imageMaxWidth": 1024,
    "imageMaxHeight": 1024,
    "loadImageFileTypes": r"/^image\/(gif|jpeg|png|bmp|svg\+xml)$/",
    "sequentialUploads": "true",
    "acceptFileTypes": r"/(\.|\/)(png|gif|bmp|jpe?g)$/i",
    "imageOrientation": True,
    "maxFileSize": 1.5 * 1024 ** 2,  # 1.5Mb
    "minFileSize": 0.0001 * 1024 ** 2,  # 0.0001Mb
    "disableImageResize": "/Android(?!.*Chrome)|Opera/.test(window.navigator "
                          "&& navigator.userAgent)",
}
"""dict: The default options for jQuery-File-Upload module.
"""


# {{{ DO NOT change this if you have deployed this app on you production server!
# Or else it might cause unexpected result!

DEFAULT_UPLOAD_URL_NAME = "galleryfield-builtingalleryimage-upload"
DEFAULT_CROP_URL_NAME = "galleryfield-builtingalleryimage-crop"
DEFAULT_FETCH_URL_NAME = "galleryfield-builtingalleryimage-fetch"

DEFAULT_TARGET_IMAGE_MODEL = "galleryfield.BuiltInGalleryImage"
DEFAULT_TARGET_IMAGE_FIELD_NAME = "image"
DEFAULT_CREATOR_FIELD_NAME = "creator"

# }}}

# BE CAUTIOUS: changing the sequence of the assets
# might result in failure to render the widget

VENDER_JS_NAMES = (
    'jquery.js',
    'jquery-ui.js',
    'jquery.ui.widget.min.js',
    'blueimp-tmpl.js',
    'load-image.all.min.js',
    'blueimp-canvas-to-blob.js',
    'bootstrap.js',
    'jquery.iframe-transport.js',
    'jquery.fileupload.js',
    'jquery.fileupload-process.js',
    'jquery.fileupload-image.js',
    'jquery.fileupload-audio.js',
    'jquery.fileupload-video.js',
    'jquery.fileupload-validate.js',
    'jquery.fileupload-ui.js',
    'jquery.blueimp-gallery.js',
    'blueimp-gallery-indicator.js',
    'blueimp-gallery-fullscreen.js',
    'cropper.js')

VENDER_CSS_NAMES = (
    'bootstrap.css',
    'jquery-ui.theme.css',
    'jquery.fileupload.css',
    'jquery.fileupload-ui.css',
    'blueimp-gallery.css',
    'blueimp-gallery-indicator.css',
    'font-awesome.css',
    'cropper.css')

DEFAULT_ASSETS = {
    # js assets
    "jquery.js": "jquery/dist/jquery.min.js",
    "jquery-ui.js": "jquery-ui-dist/jquery-ui.min.js",
    "jquery.ui.widget.min.js":
        "blueimp-file-upload/js/vendor/jquery.ui.widget.js",
    "blueimp-tmpl.js":
        "blueimp-tmpl/js/tmpl.min.js",
    "load-image.all.min.js":
        "blueimp-load-image/js/load-image.all.min.js",
    "blueimp-canvas-to-blob.js":
        "blueimp-canvas-to-blob/js/canvas-to-blob.min.js",
    "bootstrap.js":
        "bootstrap/dist/js/bootstrap.min.js",
    "jquery.iframe-transport.js":
        "blueimp-file-upload/js/jquery.iframe-transport.js",
    "jquery.fileupload.js": "blueimp-file-upload/js/jquery.fileupload.js",
    "jquery.fileupload-process.js":
        "blueimp-file-upload/js/jquery.fileupload-process.js",
    "jquery.fileupload-image.js":
        "blueimp-file-upload/js/jquery.fileupload-image.js",
    "jquery.fileupload-audio.js":
        "blueimp-file-upload/js/jquery.fileupload-audio.js",
    "jquery.fileupload-video.js":
        "blueimp-file-upload/js/jquery.fileupload-video.js",
    "jquery.fileupload-validate.js":
        "blueimp-file-upload/js/jquery.fileupload-validate.js",
    "jquery.fileupload-ui.js":
        "blueimp-file-upload/js/jquery.fileupload-ui.js",
    "jquery.blueimp-gallery.js":
        "blueimp-gallery/js/jquery.blueimp-gallery.min.js",
    "blueimp-gallery-fullscreen.js":
        "blueimp-gallery/js/blueimp-gallery-fullscreen.js",
    "blueimp-gallery-indicator.js":
        "blueimp-gallery/js/blueimp-gallery-indicator.js",
    "cropper.js": "cropper/dist/cropper.min.js",

    # css assets
    "bootstrap.css": "bootstrap/dist/css/bootstrap.min.css",
    "jquery-ui.theme.css": "jquery-ui-dist/jquery-ui.theme.min.css",
    "jquery.fileupload.css": "blueimp-file-upload/css/jquery.fileupload.css",
    "jquery.fileupload-ui.css":
        "blueimp-file-upload/css/jquery.fileupload-ui.css",
    "blueimp-gallery.css": "blueimp-gallery/css/blueimp-gallery.min.css",
    "blueimp-gallery-indicator.css":
        "blueimp-gallery/css/blueimp-gallery-indicator.css",
    "font-awesome.css": "font-awesome/css/font-awesome.min.css",
    "cropper.css": "cropper/dist/cropper.min.css",
}
"""dict: The default assets used to render the :class:`GalleryWidget` instance.
"""

BUILT_IN_JS = ["js/jquery.fileupload-ui-gallery-widget.js"]
BUILT_IN_CSS = []

EXTRA_JS = []
EXTRA_CSS = []

DEFAULT_THUMBNAIL_SIZE = "120x120"
DEFAULT_THUMBNAIL_QUALITY = 80

WIDGET_HIDDEN_INPUT_CSS_CLASS = "django-galleryfield"

PROMPT_ALERT_ON_WINDOW_RELOAD_IF_CHANGED = True



import json
import os
from io import BytesIO
from urllib.parse import unquote

from django import forms
from django.apps import apps
from django.core.exceptions import (ImproperlyConfigured, PermissionDenied,
                                    SuspiciousOperation)
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import Case, When
from django.http import JsonResponse
from django.urls import reverse
from django.utils.translation import gettext
from django.views.generic import UpdateView
from django.views.generic.list import BaseListView
from PIL import Image
from sorl.thumbnail import get_thumbnail

from galleryfield import conf, defaults
from galleryfield.utils import (get_formatted_thumbnail_size,
                                get_or_check_image_field)


class BaseImageModelMixin:
    """
    :attr:`target_model`: A valid target image model used by the view.
    :attr:`crop_url_name`: str, An URL name or an URL for handling server side
       cropping of uploaded images, if configured None, it will used the value
       in the form of ``app_label-model_name-crop`` in lower case. For example, if
       ``self.target_model`` is ``my_app.MyImage``, then the `crop_url_name`
       is auto-configured to ``my_app-myimage-crop``. You need to make sure you
       had that URL name in your URL_CONF and related views exists.

    :attr:`disable_server_side_crop`: bool, determining whether server side crop
       for this view will be enabled, defaults to True. If False, related widget
       ui won't show `Edit` buttons for uploaded images.
    """
    target_model = None
    crop_url_name = None
    disable_server_side_crop = True

    def setup(self, request, *args, **kwargs):
        # XML request only check
        if request.META.get('HTTP_X_REQUESTED_WITH') != 'XMLHttpRequest':
            raise PermissionDenied(
                gettext("Only XMLHttpRequest requests are allowed"))

        super().setup(request, *args, **kwargs)
        self.setup_model_and_image_field()
        self.validate_crop_url()
        self.thumbnail_size = self.get_and_validate_thumbnail_size_from_request()

    def setup_model_and_image_field(self):
        if self.target_model is None:
            raise ImproperlyConfigured(
                "Using BaseImageModelMixin (base class of %s) without "
                "the 'target_model' attribute is prohibited."
                % self.__class__.__name__
            )

        self._image_field_name = (get_or_check_image_field(
            obj=self, target_model=self.target_model,
            check_id_prefix=self.__class__.__name__,
            is_checking=False).name)
        self.model = apps.get_model(self.target_model)

    def validate_crop_url(self):
        if self.disable_server_side_crop:
            return
        if self.crop_url_name is None:
            app_model_name = "-".join(self.target_model.split(".")).lower()
            self.crop_url_name = "%s-crop" % app_model_name
        try:
            reverse(self.crop_url_name, kwargs={"pk": 1})
        except Exception as e:
            raise ImproperlyConfigured(
                "'crop_url_name' in %s is invalid. The exception is: "
                "%s: %s."
                % (self.__class__.__name__,
                   type(e).__name__,
                   str(e)))
        if (self.crop_url_name == defaults.DEFAULT_CROP_URL_NAME
                and self.target_model != defaults.DEFAULT_TARGET_IMAGE_MODEL):
            raise ImproperlyConfigured(
                    "'crop_url_name' in %s is using built-in default, while "
                    "'target_model' is not using built-in default value. They "
                    "are handling different image models. This is prohibited."
                    % self.__class__.__name__
            )

    def get_and_validate_thumbnail_size_from_request(self):
        # Get preview size from request
        method = self.request.method.lower()

        if method == "get":
            # If thumbnail_size is a list, request query string should be in
            # the form of &thumbnail_size=100&thumbnail_size=150, and the
            # 'get' method should be 'getlist'
            # Ref: https://stackoverflow.com/a/30107874/3437454
            thumbnail_size = self.request.GET.getlist(
                'thumbnail_size', conf.DEFAULT_THUMBNAIL_SIZE)
            error_msg = gettext(
                "'thumbnail_size' must be an int, or a string of int, "
                "or a string in the form of '80x60', or a list of"
                "2 ints, e.g, [80, 60]. "
                "Ref: https://stackoverflow.com/a/30107874/3437454"
            )
        else:
            thumbnail_size = self.request.POST.get(
                'thumbnail_size', conf.DEFAULT_THUMBNAIL_SIZE)
            error_msg = gettext(
                "'thumbnail_size' must be an int, or a string of int, "
                "or a string in the form of '80x60', or a list or tuple of "
                "2 ints, e.g, [80, 60] or (80, 60)."
            )

        try:
            return get_formatted_thumbnail_size(thumbnail_size)
        except Exception:
            raise ImproperlyConfigured(error_msg)

    def get_thumbnail(self, image):
        return get_thumbnail(
            file_=image,
            geometry_string=self.thumbnail_size,
            crop="center",
            quality=conf.DEFAULT_THUMBNAIL_QUALITY)

    def get_serialized_image(self, obj):
        # This is used to construct return value file dict in
        # upload list and crop views.
        image = getattr(obj, self._image_field_name)

        result = {
            'pk': obj.pk,
            'name': os.path.basename(image.path),
            'url': image.url,
        }

        error = []

        try:
            image_size = image.size
        except OSError:
            error.append(gettext(
                "image: The image was unexpectedly deleted from server"))
        else:
            result.update({
                "size": image_size,
            })

        if not self.disable_server_side_crop:
            result["cropUrl"] = reverse(self.crop_url_name, kwargs={"pk": obj.pk})

        try:
            result['thumbnailUrl'] = self.get_thumbnail(image).url
        except Exception as e:
            error.append(
                gettext("thumbnail: %s: %s" % (type(e).__name__, str(e)))
            )

        if error:
            result["error"] = "; ".join(error)
        return result

    def render_to_response(self, context, **response_kwargs):
        # Overriding the method from template view, we don't need
        # a template in rendering the JsonResponse
        encoder = response_kwargs.pop("encoder", DjangoJSONEncoder)
        safe = response_kwargs.pop("safe", True)
        json_dumps_params = response_kwargs.pop("json_dumps_params", None)
        return JsonResponse(
            context, encoder, safe, json_dumps_params, **response_kwargs)


class ImageFormViewMixin:
    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.form_class = self.get_form_class()

    def form_valid(self, form):
        return self.render_to_response(self.get_context_data(form=form))

    def get_context_data(self, form):
        context = {}
        if form.is_valid():
            if self.object:
                context["files"] = [self.get_serialized_image(self.object)]
                context["message"] = gettext("Done")
        else:
            context["errors"] = form.errors

        return context


class BaseCreateMixin(ImageFormViewMixin, BaseImageModelMixin):
    http_method_names = ['post']

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)

    def get_form_class(self):
        # Here we were simulating the request is done through
        # a form. In this way, we can used ImageField to validate
        # the file.

        this = self

        class ImageForm(forms.ModelForm):
            class Meta:
                model = this.model
                fields = (this._image_field_name,)

            def __init__(self, files=None, **kwargs):
                if files is not None:
                    # When getting data from ajax post
                    files = {this._image_field_name: files["files[]"]}
                return super().__init__(files=files, **kwargs)

        return ImageForm

    def form_invalid(self, form):
        """If the form is invalid, render the invalid form error."""
        return self.render_to_response(
            self.get_context_data(form=form), status=400)


class BaseListViewMixin(BaseImageModelMixin, BaseListView):
    # List view doesn't include a form

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self._pks = self.get_and_validate_pks_from_request()

    def get_and_validate_pks_from_request(self):
        # convert the request data "pks" (which is supposed to
        # be a stringfied json string) into a list of
        # image instance "pk".
        pks = self.request.GET.get("pks", None)
        if not pks:
            raise SuspiciousOperation(
                gettext("The request doesn't contain pks data"))

        try:
            pks = json.loads(unquote(pks))
            assert isinstance(pks, list)
        except Exception as e:
            raise SuspiciousOperation(
                gettext("Invalid format of pks %s: %s: %s" %
                        (str(pks), type(e).__name__, str(e))))
        else:
            for pk in pks:
                if not str(pk).isdigit():
                    raise SuspiciousOperation(
                        gettext(
                            "pks should only contain integers, while got %s"
                            % str(pk)))
        return pks

    def get_queryset(self):
        # We have to override this method, because currently super().get_queryset
        # doesn't accept Case in 'order_by'
        # See https://github.com/django/django/pull/14757
        queryset = self.model._default_manager.all()
        ordering = self.get_ordering()
        if ordering:  # pragma: no cover
            if not isinstance(ordering, (list, tuple)):
                ordering = (ordering,)
            queryset = queryset.order_by(*ordering)

        return queryset.filter(pk__in=self._pks)

    def get_ordering(self):
        # Preserving the sequence while filter by (id__in=pks)
        # See https://stackoverflow.com/a/37648265/3437454
        preserved = Case(
            *[When(pk=pk, then=pos) for pos, pk in enumerate(self._pks)])
        return preserved

    def get_context_data(self, **kwargs):
        # Return a list of serialized files
        context = {
            "files":  [self.get_serialized_image(obj)
                       for obj in self.get_queryset()]}
        context.update(kwargs)

        return context


class CropError(Exception):
    pass


class BaseCropViewMixin(ImageFormViewMixin, BaseImageModelMixin, UpdateView):
    http_method_names = ['post']

    def get_form_class(self):
        # Here we were simulating the request is done through
        # a form. In this way, we can used ImageField to validate
        # the file.
        this = self

        class ImageForm(forms.ModelForm):
            class Meta:
                model = this.model
                fields = (this._image_field_name,)

            def __init__(self, **kwargs):
                super().__init__(**kwargs)

                new_uploaded_file = this.get_cropped_uploaded_file(self.instance)
                self.instance.pk = None
                self.initial = {this._image_field_name: new_uploaded_file}

        return ImageForm

    def setup(self, request, *args, **kwargs):
        if self.disable_server_side_crop:
            raise SuspiciousOperation(
                gettext("Server side crop is not enabled."))
        super().setup(request, *args, **kwargs)
        self._cropped_result = self.get_and_validate_cropped_result_from_request()

    def get_and_validate_cropped_result_from_request(self):
        try:
            cropped_result = json.loads(self.request.POST["cropped_result"])
        except Exception as e:
            if isinstance(e, KeyError):
                raise SuspiciousOperation(
                    gettext("The request doesn't contain cropped_result"))
            raise SuspiciousOperation(
                gettext("Error while getting cropped_result: %s: %s"
                        % (type(e).__name__, str(e))))
        else:
            try:
                x = int(float(cropped_result['x']))
                y = int(float(cropped_result['y']))
                width = int(float(cropped_result['width']))
                height = int(float(cropped_result['height']))
                rotate = int(float(cropped_result['rotate']))

                # todo: allow show resized image in model ui
                try:
                    scale_x = float(cropped_result['scaleX'])
                except KeyError:
                    scale_x = None
                try:
                    scale_y = float(cropped_result['scaleY'])
                except KeyError:
                    scale_y = None
            except Exception:
                raise SuspiciousOperation(
                    gettext('Wrong format of crop_result data.'))

        return x, y, width, height, rotate, scale_x, scale_y

    def get_cropped_uploaded_file(self, old_instance):
        # This view is put into the init method of self.form_class (i.e., ImageForm)
        # to simulate we posted a NEW image in the form to create a (cropped) NEW
        # image
        old_image = getattr(old_instance, self._image_field_name)

        try:
            new_image = Image.open(old_image.path)
        except IOError:
            raise SuspiciousOperation(
                gettext('File not found，please re-upload the image'))

        image_format = new_image.format

        x, y, width, height, rotate, scale_x, scale_y = self._cropped_result

        if rotate != 0:
            # or it will raise "AttributeError: 'NoneType' object has no attribute
            # 'mode' error in pillow 3.3.0
            new_image = new_image.rotate(-rotate, expand=True)

        box = (x, y, x + width, y + height)
        new_image = new_image.crop(box)

        new_image_io = BytesIO()
        new_image.save(new_image_io, format=image_format)

        upload_file = InMemoryUploadedFile(
            file=new_image_io,
            field_name=self._image_field_name,
            name=old_image.name,
            content_type=Image.MIME[image_format],
            size=new_image_io.tell(),
            charset=None
        )
        return upload_file



from django.apps import apps
from django.conf import settings
from django.core import checks

from galleryfield import defaults
from galleryfield.utils import (GENERIC_ERROR_PATTERN, INSTANCE_ERROR_PATTERN,
                                DJGalleryCriticalCheckMessage,
                                InvalidThumbnailFormat,
                                get_formatted_thumbnail_size)

DJANGO_GALLERY_FIELD_CONFIG = "DJANGO_GALLERY_FIELD_CONFIG"

ASSETS = "assets"
EXTRA_JS = "extra_js"
EXTRA_CSS = "extra_css"

THUMBNAILS = "thumbnails"
THUMBNAIL_SIZE = "size"
THUMBNAIL_QUALITY = "quality"

WIDGET_HIDDEN_INPUT_CSS_CLASS = "widget_hidden_input_css_class"

PROMPT_ALERT_IF_CHANGED_ON_WINDOW_RELOAD = (
    "prompt_alert_if_changed_on_window_reload")

JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS = (
    "jquery_file_upload_ui_options"
)

MAX_NUMBER_OF_FILES = "maxNumberOfFiles"
SINGLE_FILE_UPLOADS = "singleFileUploads"
PREVIEW_MAX_WIDTH = "previewMaxWidth"
PREVIEW_MAX_HEIGHT = "previewMaxHeight"


def register_galleryfield_settings_checks():
    checks.register(check_settings, "django_gallery_field_checks")


def check_settings(app_configs, **kwargs):
    errors = []

    if not apps.is_installed('sorl.thumbnail'):
        missing_app = checks.Error(
            "'sorl.thumbnail' must be in INSTALLED_APPS in order "
            "to generate thumbnail for gallery images.",
            id="django-galleryfield.E001",
        )
        errors.append(missing_app)

    conf = getattr(settings, "DJANGO_GALLERY_FIELD_CONFIG", None)
    if conf is None:
        return errors

    if not isinstance(conf, dict):
        errors.append(DJGalleryCriticalCheckMessage(
            msg=(INSTANCE_ERROR_PATTERN
                 % {"location": DJANGO_GALLERY_FIELD_CONFIG, "types": "dict"}),
            id="django-galleryfield.E002"
        ))
        return errors

    assets = conf.get(ASSETS, None)
    if assets is not None:
        if not isinstance(assets, dict):
            errors.append(DJGalleryCriticalCheckMessage(
                msg=(INSTANCE_ERROR_PATTERN
                     % {"location": "'%s' in '%s'" % (
                            ASSETS, DJANGO_GALLERY_FIELD_CONFIG),
                        "types": "dict"}),
                id="django-galleryfield-assets.E001"
            ))
        else:
            assets_copy = assets.copy()
            extra_js = assets_copy.pop(EXTRA_JS, None)
            if extra_js is not None:
                if not isinstance(extra_js, list):
                    errors.append(DJGalleryCriticalCheckMessage(
                        msg=(INSTANCE_ERROR_PATTERN
                             % {"location": "'%s' in '%s' in '%s'" % (
                                    EXTRA_JS, ASSETS,
                                    DJANGO_GALLERY_FIELD_CONFIG),
                                "types": "str"}),
                        id="django-galleryfield-assets.E005"
                    ))
                else:
                    for js in extra_js:
                        if not isinstance(js, str):
                            errors.append(DJGalleryCriticalCheckMessage(
                                msg=(INSTANCE_ERROR_PATTERN % {
                                    "location":
                                        "'%s' in '%s' in '%s' in '%s'" % (
                                            str(js),
                                            EXTRA_JS, ASSETS,
                                            DJANGO_GALLERY_FIELD_CONFIG),
                                    "types": "str"}),
                                id="django-galleryfield-assets.E006"
                            ))

            extra_css = assets_copy.pop(EXTRA_CSS, None)
            if extra_css is not None:
                if not isinstance(extra_css, list):
                    errors.append(DJGalleryCriticalCheckMessage(
                        msg=(INSTANCE_ERROR_PATTERN
                             % {"location": "'%s' in '%s' in '%s'" % (
                                    EXTRA_CSS, ASSETS,
                                    DJANGO_GALLERY_FIELD_CONFIG),
                                "types": "str"}),
                        id="django-galleryfield-assets.E007"
                    ))
                else:
                    for css in extra_css:
                        if not isinstance(css, str):
                            errors.append(DJGalleryCriticalCheckMessage(
                                msg=(INSTANCE_ERROR_PATTERN % {
                                    "location":
                                        "'%s' in '%s' in '%s' in '%s'" % (
                                            str(css),
                                            EXTRA_CSS, ASSETS,
                                            DJANGO_GALLERY_FIELD_CONFIG),
                                    "types": "str"}),
                                id="django-galleryfield-assets.E008"
                            ))

            for asset_name, asset_value in assets_copy.items():
                if asset_name not in (
                        defaults.VENDER_CSS_NAMES + defaults.VENDER_JS_NAMES):
                    errors.append(checks.Warning(
                        msg=("%(location)s is not in the required assets, it"
                             "you want to use it, put it in %(extra_css)s or "
                             "%(extra_js)s"
                             % {"location": (
                                        "Asset '%s' in '%s' in '%s'" % (
                                            str(asset_name), ASSETS,
                                            DJANGO_GALLERY_FIELD_CONFIG)),
                                 "extra_js": EXTRA_JS,
                                 "extra_css": EXTRA_CSS}
                             ),
                        id="django-galleryfield-assets.W001"
                    ))
                    continue

                if asset_value is None:
                    errors.append(checks.Warning(
                        msg=("%(location)s is set to None, that might "
                             "cause unexpected result in views as well "
                             "as in admin."
                             % {"location": (
                                        "Asset '%s' in '%s' in '%s'" % (
                                            str(asset_name), ASSETS,
                                            DJANGO_GALLERY_FIELD_CONFIG))}
                             ),
                        id="django-galleryfield-assets.W002"
                    ))
                    continue

                # Note the asset_value is not necessary a str, it can be a
                # JS object (see django-js-asset).

    thumbnails = conf.get(THUMBNAILS, None)
    if thumbnails is not None:
        if not isinstance(thumbnails, dict):
            errors.append(DJGalleryCriticalCheckMessage(
                msg=(INSTANCE_ERROR_PATTERN
                     % {"location": "'%s' in '%s'" % (
                            THUMBNAILS, DJANGO_GALLERY_FIELD_CONFIG),
                        "types": "dict"}),
                id="django-galleryfield-thumbnails.E001"
            ))
        else:
            thumbnail_size = thumbnails.get(THUMBNAIL_SIZE, None)
            try:
                get_formatted_thumbnail_size(
                    thumbnail_size, name=THUMBNAIL_SIZE)
            except Exception as e:
                if isinstance(e, InvalidThumbnailFormat):
                    errors.append(DJGalleryCriticalCheckMessage(
                        msg=(GENERIC_ERROR_PATTERN
                             % {"location": "'%s' in '%s' in '%s'" % (
                                    THUMBNAIL_SIZE, THUMBNAILS,
                                    DJANGO_GALLERY_FIELD_CONFIG),
                                "error_type": type(e).__name__,
                                "error_str": (
                                        "'%s' must be an int, or a string of int, "
                                        "or in the form of 80x60, or a list, e.g,"
                                        " [80, 60], or a tuple (80, 60)"
                                        % THUMBNAIL_SIZE)}
                             ),
                        id="django-galleryfield-thumbnails.E003"
                    ))
                else:
                    errors.append(DJGalleryCriticalCheckMessage(
                        msg=(GENERIC_ERROR_PATTERN
                             % {"location": "'%s' in '%s' in '%s'" % (
                                    THUMBNAIL_SIZE, THUMBNAILS,
                                    DJANGO_GALLERY_FIELD_CONFIG),
                                "error_type": type(e).__name__,
                                "error_str": str(e)}
                             ),
                        id="django-galleryfield-thumbnails.E002"
                    ))

            thumbnail_quality = thumbnails.get(THUMBNAIL_QUALITY, None)
            if thumbnail_quality is not None:
                try:
                    quality = float(thumbnail_quality)
                except Exception as e:
                    errors.append(DJGalleryCriticalCheckMessage(
                        msg=(GENERIC_ERROR_PATTERN
                             % {"location": "'%s' in '%s' in '%s'" % (
                                    THUMBNAIL_QUALITY, THUMBNAILS,
                                    DJANGO_GALLERY_FIELD_CONFIG),
                                "error_type": type(e).__name__,
                                "error_str": str(e)}
                             ),
                        id="django-galleryfield-thumbnails.E004"
                    ))
                else:
                    if quality < 0 or quality > 100:
                        errors.append(DJGalleryCriticalCheckMessage(
                            msg=(GENERIC_ERROR_PATTERN
                                 % {"location": "'%s' in '%s' in '%s'" % (
                                        THUMBNAIL_QUALITY, THUMBNAILS,
                                        DJANGO_GALLERY_FIELD_CONFIG),
                                    "error_type": TypeError.__name__,
                                    "error_str":
                                        "Thumbnail quality should be "
                                        "between 0 and 100"}
                                 ),
                            id="django-galleryfield-thumbnails.E005"
                        ))

    jfu_options = conf.get(JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS, None)
    if jfu_options is not None:
        if not isinstance(jfu_options, dict):
            errors.append(DJGalleryCriticalCheckMessage(
                msg=(INSTANCE_ERROR_PATTERN
                     % {"location": "'%s' in '%s'" % (
                            JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS,
                            DJANGO_GALLERY_FIELD_CONFIG),
                        "types": "dict"}),
                id="django-galleryfield-jquery_file_upload_ui_options.E001"
            ))
        else:
            if MAX_NUMBER_OF_FILES in jfu_options:
                errors.append(checks.Warning(
                    msg=("%(location)s will be ignored, that is configured in "
                         "the formfield the GalleryWidget is serving, with the "
                         "param name `max_number_of_images`"
                         % {"location": (
                                    "option '%s' in '%s' in '%s'" % (
                                        MAX_NUMBER_OF_FILES,
                                        JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS,
                                        DJANGO_GALLERY_FIELD_CONFIG))}),
                    id="django-galleryfield-jquery_file_upload_ui_options.W001"
                ))

            if (SINGLE_FILE_UPLOADS in jfu_options
                    # Some user might use "false" string to represent False
                    and str(jfu_options[SINGLE_FILE_UPLOADS]).lower() == "false"):
                errors.append(checks.Warning(
                    msg=("%(location)s set to False is not allowed and will"
                         "be ignored. "
                         % {"location": (
                                    "option '%s' in '%s' in '%s'" % (
                                        SINGLE_FILE_UPLOADS,
                                        JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS,
                                        DJANGO_GALLERY_FIELD_CONFIG))}),
                    id="django-galleryfield-jquery_file_upload_ui_options.W002"
                ))

            if PREVIEW_MAX_WIDTH in jfu_options or PREVIEW_MAX_HEIGHT in jfu_options:
                errors.append(checks.Warning(
                    msg=("%(location)s will be ignored. By preview size, we mean "
                         "the thumbnail size in the GalleryWidget UI, so you should "
                         "set the value in %(right_place)s."
                         % {"location": (
                                    "option '%s' or '%s' in '%s' in '%s'" % (
                                        PREVIEW_MAX_WIDTH, PREVIEW_MAX_HEIGHT,
                                        JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS,
                                        DJANGO_GALLERY_FIELD_CONFIG)),
                             "right_place": "'%s' in '%s' in '%s'" % (
                                 THUMBNAIL_SIZE, THUMBNAILS,
                                 DJANGO_GALLERY_FIELD_CONFIG)}),
                    id="django-galleryfield-jquery_file_upload_ui_options.W003"
                ))

    widget_hidden_input_css_class = conf.get(WIDGET_HIDDEN_INPUT_CSS_CLASS, None)

    if widget_hidden_input_css_class is not None:
        if not isinstance(widget_hidden_input_css_class, str):
            errors.append(DJGalleryCriticalCheckMessage(
                msg=(INSTANCE_ERROR_PATTERN
                     % {"location": "'%s' in '%s'" % (
                            WIDGET_HIDDEN_INPUT_CSS_CLASS,
                            DJANGO_GALLERY_FIELD_CONFIG),
                        "types": "str"}),
                id="django-galleryfield-widget_hidden_input_css_class.E001"
            ))

    prompt_alert_if_changed_on_window_reload = conf.get(
        PROMPT_ALERT_IF_CHANGED_ON_WINDOW_RELOAD, None)

    if prompt_alert_if_changed_on_window_reload is not None:
        if not isinstance(prompt_alert_if_changed_on_window_reload, bool):
            errors.append(DJGalleryCriticalCheckMessage(
                msg=(INSTANCE_ERROR_PATTERN
                     % {"location": "'%s' in '%s'" % (
                            PROMPT_ALERT_IF_CHANGED_ON_WINDOW_RELOAD,
                            DJANGO_GALLERY_FIELD_CONFIG),
                        "types": "bool"}),
                id="django-galleryfield-prompt_alert_if_changed_on_window_reload.E001"  # noqa
            ))

    return errors



from django import forms
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.core.validators import BaseValidator
from django.db import models
from django.db.models import Case, IntegerField, Value, When
from django.db.models.query_utils import DeferredAttribute
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext_lazy

from galleryfield import conf
from galleryfield import defaults as _defaults
from galleryfield.utils import apps, get_or_check_image_field, logger
from galleryfield.widgets import GalleryWidget


@deconstructible
class MaxNumberOfImageValidator(BaseValidator):
    message = ngettext_lazy(
        'Number of images exceeded, only %(limit_value)d allowed',
        'Number of images exceeded, only %(limit_value)d allowed',
        'limit_value')
    code = 'max_number_of_images'

    def compare(self, a, b):
        return a > b

    def clean(self, x):
        return len(x)


class GalleryDescriptor(DeferredAttribute):
    """
    A collections of pks of image model instances
    """

    # Used django.db.models.fields.files.FileDescriptor as an example.

    def __set__(self, instance, value):
        instance.__dict__[self.field.attname] = value

    def __get__(self, instance, cls=None):
        image_list = super().__get__(instance, cls)

        if not isinstance(image_list, GalleryImages):
            attr = self.field.attr_class(instance, self.field, image_list)
            instance.__dict__[self.field.name] = attr

        return instance.__dict__[self.field.name]


class GalleryImages(list):
    def __init__(self, instance, field, field_value):
        # When field_value is None,
        # (This happens when the GalleryField was saved as null)
        field_value = field_value or []
        super().__init__(field_value)
        self._field = field
        self.instance = instance
        self._value = field_value or []

    @property
    def objects(self):
        model = apps.get_model(self._field.target_model)

        # Preserving the order of image using id__in=pks
        # https://stackoverflow.com/a/37146498/3437454
        cases = [When(id=x, then=Value(i)) for i, x in enumerate(self._value)]
        case = Case(*cases, output_field=IntegerField())
        filter_kwargs = {"id__in": self._value}
        queryset = model.objects.filter(**filter_kwargs)
        queryset = queryset.annotate(_order=case).order_by('_order')
        return queryset


class GalleryField(models.JSONField):
    """This is a model field which saves the id of images.

    :param target_model: A string in the form of ``"app_label.model_name"``,
           which can be loaded by :meth:`django.apps.get_model` (see 
           `Django docs <https://docs.djangoproject.com/en/dev/ref/applications/#django.apps.apps.get_model>`_),
           defaults to `None`. If `None`, :class:`galleryfield.BuiltInGalleryImage` will be used.
           If set, it should be :ref:`a valid image model <customize-valid-image-model>`.

    :type target_model: str, optional.

    """  # noqa

    attr_class = GalleryImages
    descriptor_class = GalleryDescriptor

    description = "A collections pks of Image"

    def contribute_to_class(self, cls, name, private_only=False):
        super().contribute_to_class(cls, name, private_only)
        setattr(cls, self.attname, self.descriptor_class(self))

    def __init__(self, target_model=None, *args, **kwargs):
        self._init_target_model = self.target_model = target_model
        if target_model is None:
            self.target_model = _defaults.DEFAULT_TARGET_IMAGE_MODEL

        self.target_model_image_field = (
            self._get_image_field_or_test(is_checking=False))

        super().__init__(*args, **kwargs)

    def _get_image_field_or_test(self, is_checking=False):
        return get_or_check_image_field(
            obj=self,
            target_model=self._init_target_model,
            check_id_prefix="gallery_field",
            is_checking=is_checking)

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        errors.extend(self._check_target_model())
        return errors

    def _check_target_model(self):
        return self._get_image_field_or_test(is_checking=True)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs['target_model'] = self.target_model
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        defaults = ({
            "required": True,

            # The following 2 params will be passed to GalleryWidget
            # to check if there're potential conflicts between
            # target_model and upload_url and fetch_url.
            # see GalleryWidget.set_and_check_urls()
            "target_model": self.target_model,
            "model_field": self.__class__.__name__
        })
        defaults.update(kwargs)
        formfield = super().formfield(**{
            'form_class': GalleryFormField,
            **defaults,
        })
        return formfield


class GalleryFormField(forms.JSONField):
    """The default formfield for :class:`galleryfield.fields.GalleryField`.

    :param max_number_of_images: Max allowed number of images, defaults
           to `None`, which means unlimited.
    :type max_number_of_images: int, optional.

    :param kwargs: Besides the options from parent class, the following were added:

           * target_model: str, a valid target image model which can be loaded by
             ``apps.get_model``. When this field is used in the model form,
             it is auto configured by the model instance.

             However, if this field is used as a non-model form field, when
             not specified, it will use the built-in default target image
             model ``galleryfield.BuiltInGalleryImage``.

           * widget: if not specified, defaults to ``GalleryWidget`` with default
             values.


    .. note:: If `target_model` not specified when initializing, an info will
       be logged to the stdout, which can be turned off by adding
       ``gallery_form_field.I001`` in ``settings.SILENCED_SYSTEM_CHECKS``.

    """

    default_error_messages = {
        'required': _("The submitted file is empty."),
        'invalid': _("The submitted images are invalid."),
    }

    def __init__(self, max_number_of_images=None, **kwargs):

        # The following 2 are used to validate GalleryWidget params
        # see GalleryWidget.defaults_checks()
        self._image_model = kwargs.pop("target_model", None)
        image_model_not_configured = False
        if self._image_model is None:
            # This happens when the formfield is used in a Non-model form
            image_model_not_configured = True
            self._image_model = _defaults.DEFAULT_TARGET_IMAGE_MODEL

        # Make sure the model is valid target image model
        if (self._image_model != _defaults.DEFAULT_TARGET_IMAGE_MODEL
                or image_model_not_configured):
            errors = get_or_check_image_field(
                obj=self,
                target_model=(
                    None if image_model_not_configured else self._image_model),
                check_id_prefix="gallery_form_field",
                is_checking=True)
            for error in errors:
                if error.is_serious():
                    raise ImproperlyConfigured(str(error))
                else:
                    if error.is_silenced():
                        continue
                    logger.info(str(error))

        # This is used for widget to identify which object the widget is servicing.
        # That information will be used when raising errors.
        self._widget_is_servicing = (
                kwargs.pop("model_field", None) or self.__class__.__name__)

        self._max_number_of_images = max_number_of_images
        super().__init__(**kwargs)

    _widget = GalleryWidget

    @property
    def widget(self):
        return self._widget

    @widget.setter
    def widget(self, value):
        # Property and setter are used to make sure the attributes will
        # be passed to new widget instance when the widget instance
        # is changed.
        setattr(value, "max_number_of_images", self.max_number_of_images)
        setattr(value, "image_model", self._image_model)
        setattr(value, "widget_is_servicing", self._widget_is_servicing)

        # Re-initialize the widget
        value.is_localized = bool(self.localize)
        value.is_required = self.required
        extra_attrs = self.widget_attrs(value) or {}
        value.attrs.update(extra_attrs)
        self._widget = value

        if not isinstance(self.widget, GalleryWidget):
            return

        # We set the upload_url and fetch_url
        # when the widget didn't specify them.
        self._set_widget_upload_url()
        self._set_widget_fetch_url()

    @property
    def _target_app_model_name(self):
        return "-".join(self._image_model.split(".")).lower()

    def _set_widget_upload_url(self):
        if self.widget.upload_url:
            return

        # Here we required a target_model should have a upload_url
        # name in url_conf in the form of app_label_model_name-upload
        # in lower case
        self.widget.upload_url = (
                "%s-upload" % self._target_app_model_name.lower())

    def _set_widget_fetch_url(self):
        if self.widget.disable_fetch or self.widget.fetch_url:
            return

        # Here we required a target_model should have a fetch_url
        # name in url_conf in the form of app_label-model_name-fetch
        # in lower case
        self.widget.fetch_url = (
                "%s-fetch" % self._target_app_model_name.lower())

    @property
    def max_number_of_images(self):
        return self._max_number_of_images

    @max_number_of_images.setter
    def max_number_of_images(self, value):
        if value is not None:
            if not str(value).isdigit():
                raise TypeError(
                    "'max_number_of_images' expects a positive integer, "
                    "got %s." % str(value))
            value = int(value)
        self._max_number_of_images = value
        self._widget.max_number_of_images = value

        if value:
            self.validators.append(
                MaxNumberOfImageValidator(int(value)))

    def widget_attrs(self, widget):
        # If BootStrap is loaded, "hiddeninput" is added by BootStrap.
        # However, we need that css class to check changes of the form,
        # so we added it manually.
        return {
            "class": " ".join(
                [conf.FILES_FIELD_CLASS_NAME, "hiddeninput"])
        }

    def to_python(self, value):
        converted = super().to_python(value)

        if converted in self.empty_values:
            return converted

        # Make sure the json is a list of pks
        if not isinstance(converted, list):
            raise ValidationError(
                self.error_messages['invalid'],
                code='invalid',
                params={'value': converted},
            )

        for _pk in converted:
            if not str(_pk).isdigit():
                raise ValidationError(
                    self.error_messages['invalid'],
                    code='invalid',
                    params={'value': converted},
                )

        # Make sure all pks exists
        image_model = apps.get_model(self._image_model)
        if (image_model.objects.filter(
                pk__in=list(map(int, converted))).count()
                != len(converted)):
            converted_copy = converted[:]
            converted = []
            for pk in converted_copy:
                if image_model.objects.filter(pk=pk).count():
                    converted.append(pk)

        return converted



from django.conf import settings

from galleryfield import defaults
from galleryfield.utils import get_formatted_thumbnail_size

"""
DJANGO_GALLERY_FIELD_CONFIG = {
    "assets": {
        "jquery.js": 'http://example.com/jquery.js',
        "bootstrap.css": "/my/local/bootstrap.css",
        "extra_js": [],
        "extra_css": [],
    },
    "thumbnails": {
        "size": 120,
        "quality": 80
    },
    "widget_hidden_input_css_class": "django-galleryfield",
    "prompt_alert_if_changed_on_window_reload": True,
    "jquery_file_upload_ui_options": {}
}
"""


_APP_CONFIG = getattr(settings, "DJANGO_GALLERY_FIELD_CONFIG", {})

# {{{ Generating js and css assets for GalleryWidget

_APP_CONFIG_ASSETS = _APP_CONFIG.get("assets", {})
EXTRA_JS = _APP_CONFIG_ASSETS.pop("extra_js", defaults.EXTRA_JS)
EXTRA_CSS = _APP_CONFIG_ASSETS.pop("extra_css", defaults.EXTRA_CSS)

_assets = defaults.DEFAULT_ASSETS.copy()
_assets.update(_APP_CONFIG_ASSETS)

_js = [_assets.get(v) for v in defaults.VENDER_JS_NAMES]
_js.extend(defaults.BUILT_IN_JS + EXTRA_JS)
JS = [j for j in _js if j]

_css = [_assets.get(v) for v in defaults.VENDER_CSS_NAMES] + EXTRA_CSS
_css.extend(defaults.BUILT_IN_CSS + EXTRA_CSS)
CSS = [c for c in _css if c]

# }}}

_APP_CONFIG_THUMBNAILS = _APP_CONFIG.get("thumbnails", {})
DEFAULT_THUMBNAIL_SIZE = get_formatted_thumbnail_size(
    _APP_CONFIG_THUMBNAILS.get(
        "size", defaults.DEFAULT_THUMBNAIL_SIZE))
DEFAULT_THUMBNAIL_QUALITY = int(_APP_CONFIG_THUMBNAILS.get(
    "quality", defaults.DEFAULT_THUMBNAIL_QUALITY))


_APP_CONFIG_WIDGET_INPUT_CSS_CLASS = _APP_CONFIG.get(
    "widget_hidden_input_css_class", defaults.WIDGET_HIDDEN_INPUT_CSS_CLASS)
FILES_FIELD_CLASS_NAME = _APP_CONFIG_WIDGET_INPUT_CSS_CLASS + "-files-field"


PROMPT_ALERT_ON_WINDOW_RELOAD_IF_CHANGED = _APP_CONFIG.get(
    'prompt_alert_if_changed_on_window_reload',
    defaults.PROMPT_ALERT_ON_WINDOW_RELOAD_IF_CHANGED
)

# todo: allow setting default jquery_fileupload_ui_options in settings.

JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS = _APP_CONFIG.get(
    "jquery_file_upload_ui_options",
    defaults.JQUERY_FILE_UPLOAD_UI_DEFAULT_OPTIONS
)



from django.apps import AppConfig


class GalleryConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'galleryfield'

    def ready(self):
        from .checks import register_galleryfield_settings_checks
        register_galleryfield_settings_checks()



# Generated by Django 3.2.5 on 2021-08-03 04:53

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import galleryfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='MyImageModel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('photo', models.ImageField(upload_to='')),
            ],
        ),
        migrations.CreateModel(
            name='DemoGallery',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('images', galleryfield.fields.GalleryField(blank=True, null=True, target_model='galleryfield.BuiltInGalleryImage', verbose_name='Photos')),
                ('owner', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='Owner')),
            ],
        ),
    ]



import os

from setuptools import find_packages, setup

import galleryfield


def read(fname):
    # file read function copied from sorl.django-documents project
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


install_requires = [
    'sorl-thumbnail',
    'pillow'
]

setup(
    name='django-galleryfield',
    version=galleryfield.__version__,
    description="Django AJAX upload widget and model field for multiple images, featuring drag & drop uploading, upload progress bar, sortable and croppable image gallery",
    long_description_content_type="text/x-rst",
    long_description=read('README.rst'),
    classifiers=[
        # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 3.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='django ajax html5 upload widget images gallery sorting crop progress'
             'thumbnails multiple GalleryField admin forms field',
    author='Dong Zhuang',
    author_email='dzhuang.scut@gmail.com',
    url='https://github.com/dzhuang/django-galleryfield',
    license='MIT',
    packages=find_packages(exclude=['tests', 'demo']),
    include_package_data=True,
    zip_safe=True,
    install_requires=install_requires,
    entry_points="""
        # -*- Entry points: -*-
    """,
)
